from uib_inf100_graphics import *
import random

################
#### Oppstart
################

def app_started(app):
    app.margin = 20
    app.display_info = False
    app.paused = False
    app.timer_delay = 1000//30  # 30 steg per sek ~ 1000//30 ms mellom steg
    init(app)

def init(app):
    """ Initialise the values of the model. Can be used to reset the 
    simulation. The model consists of agents moving randomly around
    on an 1x1 area. If an agent which is sick comes within app.radius
    of an agent which is healthy but not immune, it will become sick. 
    A sick agent will stay sick for app.recovery_time number of steps,
    before then becoming immune.
    """
    app.radius = 0.012          # Radius for en agent
    app.population = 200        # Antall agenter i simuleringen
    app.mobility = 0.01         # Hvor mye en agent flytter seg per steg
    app.recovery_time = 30 * 8  # Hvor lenge (antall steg) man er syk (8sek)

    app.agents = [generate_random_agent(app) for _ in range(app.population)]
    app.agents[0]["status"] = "sick" # Gjør agent 0 syk i begynnelsen
    app.walls = [
        # Murer/vegger defineres på formen (x1, y1, x2, y2) i 1x1 -området
        (0.5-app.radius, 0, 0.5+app.radius, 0.85),
    ]

    # Opptellingen av hvor mange som er syke og hvor mange som er friske
    #   Vi teller én gang i hvert steg, så app.tallies[1000] vil gi oss info
    #   om hvor mange som var friske, syke og immune i steg 1000.
    app.tallies = [{"healthy": app.population-1, "sick": 1, "immune": 0}]

def generate_random_agent(app):
    """ Generates a random and healthy agent. The agent will be placed
    in a random location in the 1x1 area. The agent will be located such
    that the distances to any boarder is at least app.radius."""
    # Et oppslagsverk benyttes for å representerer en agent
    return {
        "status": "healthy", # healthy, sick eller immune
        "steps_since_status_change": 0,
        # x og y -koordinatene genereres slik at de er helt innenfor brettet
        # (med en avstand på minst app.radius til brettets kant)
        "x": app.radius + random.random()*(1 - app.radius*2),
        "y": app.radius + random.random()*(1 - app.radius*2),
        # dx og dy er hastigheten og retningen en agent flytter seg
        "dx": (random.random() - 0.5) * app.mobility,
        "dy": (random.random() - 0.5) * app.mobility,
    }

################
#### Kontrollere
################

def key_pressed(app, event):
    if event.key == "r":
        init(app)
    elif event.key == "p":
        app.paused = not app.paused
    elif event.key == "i":
        app.display_info = not app.display_info
    elif app.paused and event.key == "Space":
        do_step(app)

def timer_fired(app):
    if app.paused:
        return
    do_step(app)

def do_step(app):
    """ Does a single step of the simulation. This includes moving and
    healing people, making sick people infect nearby healthy people, 
    as well as performing a tally. """
    for agent in app.agents:
        move_agent(agent)
        bounce_with_walls(app, agent)
        recovery_process(agent, app.recovery_time)
        infect_others_if_sick(app, agent)

    tally_up_status(app)

def move_agent(agent):
    """ Moves a agent one step. Bounces off the edges of the 1x1 area. """
    # Flytt agenten
    agent["x"] += agent["dx"]
    agent["y"] += agent["dy"]
    # Hold agenten inne på brettet
    if agent["x"] < 0 or agent["x"] > 1:
        agent["x"] = min(1, max(0, agent["x"]))
        agent["dx"] *= -1
    if agent["y"] < 0 or agent["y"] > 1:
        agent["y"] = min(1, max(0, agent["y"]))
        agent["dy"] *= -1

def point_in_rectangle(x1, y1, x2, y2, x, y):
    """ Returns True if the point (x, y) is within the bounds of the
    rectangle (x1, y1, x2, y2) and False otherwise. If the point is
    exactly on the border, the function returns True. """
    return (min(x1, x2) <= x <= max(x1, x2) 
        and min(y1, y2) <= y <= max(y1, y2))

def bounce_with_walls(app, agent):
    """ When the agent hits a wall, it bounces. """
    px = agent["x"]
    py = agent["y"]
    r = app.radius

    for wall in app.walls:
        wx1, wy1, wx2, wy2 = wall
        # Forenklet kollisjonsdeteksjon (litt funky på hjørner)
        # TODO: Håndtere tilfeller der agenten treffer hjørnet.
        if point_in_rectangle(wx1 - r, wy1, wx2 + r, wy2, px, py):
            agent["dx"] *= -1
            agent["x"] = wx2 + r if agent["dx"] > 0 else wx1 - r
        elif point_in_rectangle(wx1, wy1 - r, wx2, wy2 + r, px, py):
            agent["dy"] *= -1
            agent["y"] = wy2 + r if agent["dy"] > 0 else wy1 - r

def recovery_process(agent, recovery_time):
    """ Handles the recovery process of a sick agent. """
    if agent["status"] == "sick":
        agent["steps_since_status_change"] += 1
        if agent["steps_since_status_change"] > recovery_time:
            agent["status"] = "immune"

def infect_others_if_sick(app, agent):
    """ If the agent is sick, let all healthy agents within 
    infection radius of the agent become sick. """
    if agent["status"] == "sick":
        for other_agent in app.agents:
            if other_agent["status"] != "healthy":
                continue
            if within_infection_radius(agent, other_agent, app.radius):
                other_agent["status"] = "sick"
                other_agent["steps_since_status_change"] = 0

def within_infection_radius(agent1, agent2, radius):
    """ Returns True if the distance between agent1 and agent2 is
    at most the given radius, otherwise it returns False. """
    distance_squared = ((agent1["x"] - agent2["x"])**2 
                      + (agent1["y"] - agent2["y"])**2)
    return distance_squared <= radius**2

def tally_up_status(app):
    """ Counts how many agents are healthy, sick and immune,
    and returns the result as a dictionary with counts. """
    if app.tallies[-1]["sick"] == 0:
        # No need to make another tally if nobody is sick anymore
        return
    tally = {"healthy": 0, "sick": 0, "immune": 0}
    for agent in app.agents:
        tally[agent["status"]] += 1
    app.tallies.append(tally)


################
#### Visning
################

def redraw_all(app, canvas):
    draw_top_message(app, canvas, app.width/2, 20)
    draw_proportion_graph(app, canvas, app.margin, app.margin + 20,
                          app.width-app.margin, app.height/3 - app.margin/2)
    draw_people_board(app, canvas, app.margin, app.height/3 + app.margin/2,
                      app.width-app.margin, app.height-app.margin)

def draw_top_message(app, canvas, x, y):
    msg = ""
    if app.display_info:
        msg = (f"population={app.population}  "
              + f"mobility={app.mobility}  "
              + f"recovery_time={app.recovery_time/30}s  "
              + f"paused={app.paused}")
    else:
        msg = "(r)estart    (p)ause    (i)nfo"
    canvas.create_text(x, y, text=msg)


def draw_proportion_graph(app, canvas, x1, y1, x2, y2):
    """ Draws the proportion graph within the bounding rectangle
    (x1, y1, x2, y2). For performance reasons, if there are more
    than 100 tallies in app.tallies, an evenly distributed sample
    of 100 tallies will be chosen instead of drawing all tallies. """

    # Filter out 100 evenly distributed tallies
    n = min(100, len(app.tallies))
    for i in range(n):
        tally = app.tallies[i * len(app.tallies) // n]
        bar_x_left = (x1 * (n - i) + x2 * i)/n
        bar_x_right = (x1 * (n - i - 1) + x2 * (i + 1))/n
        draw_bar(canvas, bar_x_left, y1, bar_x_right, y2, tally)

    canvas.create_rectangle(x1, y1, x2, y2) # Bounding rectangle

def draw_bar(canvas, x1, y1, x2, y2, tally):
    """ Draws a single data point/bar in the proportion graph. """
    total = sum(tally.values())
    height = y2 - y1
    y_top = y1
    for status in ["immune", "healthy", "sick"]:
        y_bottom = y_top + height * tally[status] / total
        canvas.create_rectangle(x1, y_top, x2, y_bottom,
                                fill=get_color_from_status(status), width=0)
        y_top = y_bottom
        
def draw_people_board(app, canvas, x1, y1, x2, y2):
    """ Draw the 1x1 board with the agents and walls within the 
    bounding rectangle given by (x1, y1, x2, y2). """
    canvas.create_rectangle(x1, y1, x2, y2)
    width = x2 - x1
    height = y2 - y1
    radius = app.radius * min(width, height)
    for agent in app.agents:
        px1 = x1 + agent["x"] * width - radius
        py1 = y1 + agent["y"] * height - radius
        px2 = px1 + 2 * radius
        py2 = py1 + 2 * radius
        color = get_color_from_status(agent["status"])
        canvas.create_oval(px1, py1, px2, py2, fill=color, width=0)

    for wx1, wy1, wx2, wy2 in app.walls:
        canvas.create_rectangle(x1 + wx1 * width, y1 + wy1 * height,
                                x1 + wx2 * width, y1 + wy2 * height,
                                fill="black", width=0)

def get_color_from_status(status):
    if status == "healthy":  return "green"
    elif status == "sick":   return "orange"
    elif status == "immune": return "magenta"
    else: return "black"

run_app(width=600, height=700)